/*
 * feedbackd
 *
 * Copyright (c) 2012 - 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <dlfcn.h>
#include <assert.h>
#include <vconf.h>
#include <time.h>
#include <string.h>
#include <libgdbus/dbus-system.h>

#include "core/log.h"
#include "core/list.h"
#include "core/common.h"
#include "core/device-idler.h"
#include "core/config-parser.h"
#include "haptic.h"

#define HAPTIC_CONF_PATH			"/etc/feedbackd/haptic.conf"
#define SIGNAL_CHANGE_HARDKEY           "ChangeHardkey"
#define SIGNAL_POWEROFF_STATE           "ChangeState"
#define SIGNAL_VIBRATOR_INITIATED       "InitiateVibrator"

/* hardkey vibration variable */
#define HARDKEY_VIB_ITERATION		1
#define HARDKEY_VIB_FEEDBACK		3
#define HARDKEY_VIB_PRIORITY		2
#define HARDKEY_VIB_DURATION		30
#define HAPTIC_FEEDBACK_STEP		20
#define DEFAULT_FEEDBACK_LEVEL		3

/* power on, power off vibration variable */
#define POWER_ON_VIB_DURATION			300
#define POWER_OFF_VIB_DURATION			300
#define POWER_VIB_FEEDBACK			100

#define MAX_EFFECT_BUFFER			(64*1024)

#ifndef VCONFKEY_RECORDER_STATE
#define VCONFKEY_RECORDER_STATE "memory/recorder/state"
#define VCONFKEY_RECORDER_STATE_RECORDING	2
#endif

#define CHECK_VALID_OPS(ops, r)		((ops) ? true : !(r = -ENODEV))
#define RETRY_CNT	3

struct haptic_info {
	char *sender;
	dd_list *handle_list;
	guint id_watch;
};

struct vibrate_effect_info {
	unsigned int handle;
	char *pattern;
	int level;
	int priority;
};

/* for playing */
static int g_handle;

/* haptic operation variable */
static dd_list *h_head;
static dd_list *haptic_handle_list;
static const struct haptic_plugin_ops *h_ops;
static enum haptic_type h_type;
static bool haptic_disabled;

struct haptic_config {
	int level;
	int *level_arr;
	int sound_capture;
};

static struct haptic_config haptic_conf;

static int haptic_start(void);
static int haptic_stop(void);
static int haptic_internal_init(void);
static int remove_haptic_info(struct haptic_info *info);

struct haptic_data cur_h_data;

void add_haptic(const struct haptic_ops *ops)
{
	DD_LIST_APPEND(h_head, (void *)ops);
}

void remove_haptic(const struct haptic_ops *ops)
{
	DD_LIST_REMOVE(h_head, (void *)ops);
}

static int haptic_module_load(void)
{
	struct haptic_ops *ops;
	dd_list *elem;
	int r;

	/* find valid plugin */
	DD_LIST_FOREACH(h_head, elem, ops) {
		if (ops->is_valid && ops->is_valid()) {
			if (ops->load)
				h_ops = ops->load();
			h_type = ops->type;
			break;
		}
	}

	if (!CHECK_VALID_OPS(h_ops, r)) {
		_E("Can't find the valid haptic device");
		return r;
	}

	/* solution bug
	 * we do not use internal vibration except power off.
	 * if the last handle is closed during the playing of vibration,
	 * solution makes unlimited vibration.
	 * so we need at least one handle. */
	haptic_internal_init();

	return 0;
}

static int convert_magnitude_by_conf(int level)
{
	int i, step;

	assert(level >= 0 && level <= 100);

	step = 100 / (haptic_conf.level-1);
	for (i = 0; i < haptic_conf.level; ++i) {
		if (level <= i*step) {
			_D("level changed : %d -> %d", level, haptic_conf.level_arr[i]);
			return haptic_conf.level_arr[i];
		}
	}

	_D("play default level");
	return DEFAULT_FEEDBACK_LEVEL * HAPTIC_FEEDBACK_STEP;
}

GVariant *hdbus_get_count(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)

{
	int ret, val;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	ret = h_ops->get_device_count(&val);
	if (ret >= 0)
		ret = val;

exit:
	return g_variant_new("(i)", ret);
}

void haptic_name_owner_changed(GDBusConnection *connection,
	const gchar     *name,
	gpointer         user_data)
{
	dd_list *n;
	struct haptic_info *info = user_data;
	int handle;

	_I("%s (sender:%s)", __func__, name);

	if (!info)
		return;

	for (n = info->handle_list; n; n = n->next) {
		handle = (int)(long)n->data;
		h_ops->stop_device(handle);
		h_ops->close_device(handle);
	}

	remove_haptic_info(info);
}

static struct haptic_info *add_haptic_info(const char *sender)
{
	struct haptic_info *info;

	assert(sender);

	info = calloc(1, sizeof(struct haptic_info));
	if (!info)
		return NULL;

	info->sender = strdup(sender);
	DD_LIST_APPEND(haptic_handle_list, info);

	info->id_watch = dbus_handle_watch_name(sender, NULL, haptic_name_owner_changed, info, NULL);

	return info;
}

static int remove_haptic_info(struct haptic_info *info)
{
	assert(info);

	dbus_handle_unwatch_name(info->id_watch);

	DD_LIST_REMOVE(haptic_handle_list, info);
	DD_LIST_FREE_LIST(info->handle_list);
	if (info->sender) {
		free(info->sender);
	}
	free(info);

	return 0;
}

static struct haptic_info *get_matched_haptic_info(const char *sender)
{
	dd_list *n;
	struct haptic_info *info;
	int len;

	assert(sender);

	len = strlen(sender) + 1;
	DD_LIST_FOREACH(haptic_handle_list, n, info) {
		if (!strncmp(info->sender, sender, len))
			return info;
	}

	return NULL;
}

GVariant *hdbus_open_device(GDBusConnection *conn,
							const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
							GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)

{
	int index, handle, ret;
	struct haptic_info *info;

	/* Load haptic module before booting done */
	if (!CHECK_VALID_OPS(h_ops, ret)) {
		ret = haptic_module_load();
		if (ret < 0)
			goto exit;
	}

	g_variant_get(param, "(i)", &index);

	ret = h_ops->open_device(index, &handle);
	if (ret < 0)
		goto exit;

	if (!sender) {
		ret = -EPERM;
		h_ops->close_device(handle);
		goto exit;
	}

	info = get_matched_haptic_info(sender);
	if (!info) {
		info = add_haptic_info(sender);
		if (!info) {
			_E("fail to create haptic information");
			ret = -EPERM;
			h_ops->close_device(handle);
			goto exit;
		}
	}

	DD_LIST_APPEND(info->handle_list, (gpointer)(long)handle);

	ret = handle;

exit:
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_close_device(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int handle;
	int ret;
	struct haptic_info *info;
	int cnt;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	g_variant_get(param, "(u)", &handle);

	if (!sender) {
		_E("fail to get sender from dbus message");
		ret = -EPERM;
		goto exit;
	}

	ret = h_ops->close_device(handle);
	if (ret < 0)
		goto exit;

	info = get_matched_haptic_info(sender);
	if (!info) {
		_E("fail to find the matched haptic info.");
		goto exit;
	}

	DD_LIST_REMOVE(info->handle_list, (gpointer)(long)handle);
	cnt = DD_LIST_LENGTH(info->handle_list);
	if (cnt == 0)
		remove_haptic_info(info);

exit:
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_vibrate_monotone(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int handle;
	int duration, level, priority, e_handle, ret = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	if (haptic_disabled)
		goto exit;

	g_variant_get(param, "(uiii)", &handle, &duration, &level, &priority);

	/* convert as per conf value */
	level = convert_magnitude_by_conf(level);
	if (level < 0) {
		ret = -EINVAL;
		goto exit;
	}
	if (priority < PRIORITY_MIN)
		priority = PRIORITY_MIN;
	else if (priority > PRIORITY_TOP)
		priority = PRIORITY_TOP;

	ret = h_ops->vibrate_monotone(handle, duration, level, priority, &e_handle);
	if (ret >= 0)
		ret = e_handle;

exit:
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_vibrate_buffer(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int handle;
	const unsigned char *data = NULL;
	GVariant *pvar = NULL;
	gsize size = 0;
	int iteration, level, priority, e_handle, ret = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	if (haptic_disabled)
		goto exit;

	g_variant_get(param, "(u@ayiii)", &handle, &pvar, &iteration, &level, &priority);
	data = g_variant_get_fixed_array(pvar, &size, sizeof(char));
	g_variant_unref(pvar);

	/* convert as per conf value */
	level = convert_magnitude_by_conf(level);
	if (level < 0) {
		ret = -EINVAL;
		goto exit;
	}

	ret = h_ops->vibrate_buffer(handle, data, iteration, level, priority, &e_handle);
	if (ret >= 0)
		ret = e_handle;

exit:
	return g_variant_new("(i)", ret);
}

static void vibrate_effect_idler_cb(void *data)
{
	struct vibrate_effect_info *vibrate_info = (struct vibrate_effect_info *)data;

	h_ops->vibrate_effect(vibrate_info->handle, vibrate_info->pattern,
			vibrate_info->level, vibrate_info->priority);
	free(vibrate_info->pattern);
	free(vibrate_info);
}

GVariant *hdbus_vibrate_effect(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	struct vibrate_effect_info *vibrate_info;
	unsigned int handle;
	char *pattern = NULL;
	int level, priority, ret = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	if (haptic_disabled)
		goto exit;

	g_variant_get(param, "(usii)",  &handle, &pattern, &level, &priority);

	/* convert as per conf value */
	level = convert_magnitude_by_conf(level);
	if (level < 0) {
		ret = -EINVAL;
		goto exit;
	}
	if (priority < PRIORITY_MIN)
		priority = PRIORITY_MIN;
	else if (priority > PRIORITY_TOP)
		priority = PRIORITY_TOP;

	vibrate_info = calloc(1, sizeof(struct vibrate_effect_info));
	if (!vibrate_info) {
		_E("failed to allocate memory for vibrate_info");
		ret = -errno;
		goto exit;
	}
	vibrate_info->handle = handle;
	vibrate_info->pattern = pattern;
	pattern = NULL;
	if (!vibrate_info->pattern) {
		_E("failed to allocate memory for pattern");
		ret = -errno;
		free(vibrate_info);
		goto exit;
	}
	vibrate_info->level = level;
	vibrate_info->priority = priority;

	ret = add_idle_request(vibrate_effect_idler_cb, (void *)vibrate_info);

exit:
	g_free(pattern);
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_stop_device(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int handle;
	int ret = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	if (haptic_disabled)
		goto exit;

	g_variant_get(param, "(u)", &handle);

	ret = h_ops->stop_device(handle);

exit:
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_get_state(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int index, state, ret;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	g_variant_get(param, "(i)", &index);

	ret = h_ops->get_device_state(index, &state);
	if (ret >= 0)
		ret = state;

exit:
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_create_effect(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	static unsigned char data[MAX_EFFECT_BUFFER];
	//static unsigned char *p = data;
	const haptic_module_effect_element *elem_arr;
	int i, cnt, ret, bufsize = sizeof(data);
	GVariant *pvar = NULL;
	gsize size = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	g_variant_get(param, "(i@ayi)", &bufsize, &pvar, &cnt);
	elem_arr = g_variant_get_fixed_array(pvar, &size, sizeof(char));
	g_variant_unref(pvar);

	if (bufsize > MAX_EFFECT_BUFFER) {
		ret = -ENOMEM;
		goto exit;
	}

	for (i = 0; i < cnt; ++i)
		_D("[%2d] %d %d", i, elem_arr[i].haptic_duration, elem_arr[i].haptic_level);

	memset(data, 0, MAX_EFFECT_BUFFER);
	ret = h_ops->create_effect(data, bufsize, elem_arr, cnt);

exit:
	pvar = g_variant_new_fixed_array(G_VARIANT_TYPE_BYTE, data, sizeof(data), sizeof(char));
	return g_variant_new("(@ayi)", pvar, ret);
}

GVariant *hdbus_get_duration(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int handle;
	const unsigned char *data;
	int duration, ret;
	GVariant *pvar = NULL;
	gsize size = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	g_variant_get(param, "(u@ay)", &handle, &pvar);
	data = g_variant_get_fixed_array(pvar, &size, sizeof(char));
	g_variant_unref(pvar);

	ret = h_ops->get_buffer_duration(handle, data, &duration);
	if (ret >= 0)
		ret = duration;

exit:
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_save_binary(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)

{
	const unsigned char *data = NULL;
	char *file_path = NULL;
	int ret;
	GVariant *pvar;
	gsize size = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	g_variant_get(param, "(@ays)", &pvar, &file_path);
	data = g_variant_get_fixed_array(pvar, &size, sizeof(char));
	g_variant_unref(pvar);

	_D("file path : %s", file_path);
	ret = h_ops->convert_binary(data, size, file_path);

exit:
	g_free(file_path);
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_show_handle_list(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	dd_list *n;
	dd_list *elem;
	struct haptic_info *info;

	_D("sender    handle");
	DD_LIST_FOREACH(haptic_handle_list, n, info) {
		for (elem = info->handle_list; elem; elem = elem->next)
			_D("%s    %d", info->sender, (int)(long)elem->data);
	}

	return g_variant_new_tuple(NULL, 0);
}

GVariant *hdbus_pattern_is_supported(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)

{
	char *data = NULL;
	int ret = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	if (haptic_disabled)
		goto exit;

	g_variant_get(param, "(s)", &data);

	ret = h_ops->is_supported(data);

	_I("%s is supported : %d", data, ret);

exit:
	g_free(data);
	return g_variant_new("(i)", ret);
}

static int haptic_internal_init(void)
{
	int r;
	if (!CHECK_VALID_OPS(h_ops, r))
		return r;
	return h_ops->open_device(HAPTIC_MODULE_DEVICE_ALL, &g_handle);
}

static int haptic_internal_exit(void)
{
	int r;
	if (!CHECK_VALID_OPS(h_ops, r))
		return r;
	return h_ops->close_device(g_handle);
}

static void haptic_hardkey_changed_cb(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	int level, status, e_handle, ret;

	if (!CHECK_VALID_OPS(h_ops, ret)) {
		ret = haptic_module_load();
		if (ret < 0)
			return;
	}

	if (!g_handle)
		haptic_internal_init();

	/* if haptic is stopped, do not play vibration */
	if (haptic_disabled)
		return;

	if (vconf_get_bool(VCONFKEY_SETAPPL_VIBRATION_STATUS_BOOL, &status) < 0) {
		_E("fail to get VCONFKEY_SETAPPL_VIBRATION_STATUS_BOOL");
		status = 1;
	}

	/* when turn off haptic feedback option */
	if (!status)
		return;

	ret = vconf_get_int(VCONFKEY_SETAPPL_TOUCH_FEEDBACK_VIBRATION_LEVEL_INT, &level);
	if (ret < 0) {
		_E("fail to get VCONFKEY_SETAPPL_TOUCH_FEEDBACK_VIBRATION_LEVEL_INT");
		level = HARDKEY_VIB_FEEDBACK;
	}

	ret = h_ops->vibrate_monotone(g_handle, HARDKEY_VIB_DURATION,
			level*HAPTIC_FEEDBACK_STEP, HARDKEY_VIB_PRIORITY, &e_handle);
	if (ret < 0)
		_E("fail to vibrate buffer : %d", ret);

	return;
}

static void haptic_poweroff_cb(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	int e_handle, ret;
	struct timespec time = {0,};

	if (!CHECK_VALID_OPS(h_ops, ret)) {
		ret = haptic_module_load();
		if (ret < 0)
			return;
	}

	if (!g_handle)
		haptic_internal_init();

	/* power off vibration */
	ret = h_ops->vibrate_monotone(g_handle, POWER_OFF_VIB_DURATION,
			POWER_VIB_FEEDBACK, HARDKEY_VIB_PRIORITY, &e_handle);
	if (ret < 0) {
		_E("fail to vibrate_monotone : %d", ret);
		return;
	}

	/* sleep for vibration */
	time.tv_nsec = POWER_OFF_VIB_DURATION * NANO_SECOND_MULTIPLIER;
	nanosleep(&time, NULL);
	return;
}

static void sound_capturing_cb(keynode_t *key, void *data)
{
	int status;

	status = vconf_keynode_get_int(key);

	/* if sound capture is in use, this value is 1(true). */
	if (status == VCONFKEY_RECORDER_STATE_RECORDING)
		haptic_stop();
	else
		haptic_start();
}

static int parse_section(struct parse_result *result, void *user_data, int index)
{
	struct haptic_config *conf = (struct haptic_config *)user_data;

	if (!result)
		return 0;

	if (!result->section || !result->name || !result->value)
		return 0;

	if (MATCH(result->name, "sound_capture")) {
		conf->sound_capture = atoi(result->value);
	} else if (MATCH(result->name, "level")) {
		conf->level = atoi(result->value);
		if (conf->level < 0 || conf->level >= INT_MAX - 1) {
			_E("You must set level with positive number in integer range");
			return -EINVAL;
		}
		conf->level_arr = calloc(sizeof(int), conf->level);
		if (!conf->level_arr) {
			_E("failed to allocate memory for level");
			return -errno;
		}
	} else if (MATCH(result->name, "value")) {
		if (index < 0)
			return -EINVAL;
		conf->level_arr[index] = atoi(result->value);
	}

	return 0;
}

static int haptic_load_config(struct parse_result *result, void *user_data)
{
	struct haptic_config *conf = (struct haptic_config *)user_data;
	char name[NAME_MAX];
	int ret;
	static int index;

	if (!result)
		return 0;

	if (!result->section || !result->name || !result->value)
		return 0;

	/* Parsing 'Haptic' section */
	if (MATCH(result->section, "Haptic")) {
		ret = parse_section(result, user_data, -1);
		if (ret < 0) {
			_E("failed to parse [Haptic] section : %d", ret);
			return ret;
		}
		goto out;
	}

	/* Parsing 'Level' section */
	for (index = 0; index < conf->level; ++index) {
		snprintf(name, sizeof(name), "level%d", index);
		if (MATCH(result->section, name)) {
			ret = parse_section(result, user_data, index);
			if (ret < 0) {
				_E("failed to parse [level] section : %d", ret);
				return ret;
			}
			goto out;
		}
	}

out:
	return 0;
}

static const dbus_method_s hdbus_methods[] = {
	{ "GetCount",          NULL,   "i", hdbus_get_count },
	{ "OpenDevice",         "i",   "i", hdbus_open_device },
	{ "CloseDevice",        "u",   "i", hdbus_close_device },
	{ "StopDevice",         "u",   "i", hdbus_stop_device },
	{ "VibrateMonotone", "uiii",   "i", hdbus_vibrate_monotone },
	{ "VibrateBuffer", "uayiii",   "i", hdbus_vibrate_buffer },
	{ "VibrateEffect",   "usii",   "i", hdbus_vibrate_effect },
	{ "GetState",           "i",   "i", hdbus_get_state },
	{ "GetDuration",      "uay",   "i", hdbus_get_duration },
	{ "CreateEffect",    "iayi", "ayi", hdbus_create_effect },
	{ "SaveBinary",       "ays",   "i", hdbus_save_binary },
	{ "ShowHandleList",    NULL,  NULL, hdbus_show_handle_list },
	{ "IsSupported",        "s",   "i", hdbus_pattern_is_supported },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.name = VIBRATOR_INTERFACE_HAPTIC,
	.methods = hdbus_methods,
	.nr_methods = ARRAY_SIZE(hdbus_methods),
};

int haptic_probe(void)
{
	/**
	 * load haptic module.
	 * if there is no haptic module,
	 * feedbackd does not activate a haptic interface.
	 */
	return haptic_module_load();
}

guint id_sig_change_hardkey;
guint id_sig_pwr_off_state;

void haptic_init(void)
{
	int r;

	/* get haptic data from configuration file */
	r = config_parse(HAPTIC_CONF_PATH, haptic_load_config, &haptic_conf);
	if (r < 0) {
		_E("failed to load configuration file(%s) : %d", HAPTIC_CONF_PATH, r);
		safe_free(haptic_conf.level_arr);
	}

	/* init dbus interface */
	r = dbus_handle_register_dbus_object(NULL, VIBRATOR_PATH_HAPTIC, &dbus_interface);
	if (r < 0)
		_E("fail to init hdbus interface and method(%d)", r);

	/* register notifier for below each event */
	id_sig_change_hardkey = subscribe_dbus_signal(NULL, DEVICED_PATH_KEY, DEVICED_INTERFACE_KEY, SIGNAL_CHANGE_HARDKEY, haptic_hardkey_changed_cb, NULL, NULL);
	if (id_sig_change_hardkey <= 0) {
		_E("Failed to register signal handler! %d", r);
		return;
	}

	id_sig_pwr_off_state = subscribe_dbus_signal(NULL, DEVICED_PATH_POWEROFF, DEVICED_INTERFACE_POWEROFF, SIGNAL_POWEROFF_STATE, haptic_poweroff_cb, NULL, NULL);
	if (id_sig_pwr_off_state <= 0) {
		_E("Failed to register signal handler! %d", r);
		return;
	}

	dbus_handle_broadcast_dbus_signal(VIBRATOR_PATH_HAPTIC,
			VIBRATOR_INTERFACE_HAPTIC,
			SIGNAL_VIBRATOR_INITIATED,
			"", NULL);

	/* add watch for sound capturing value */
	if (haptic_conf.sound_capture)
		vconf_notify_key_changed(VCONFKEY_RECORDER_STATE, sound_capturing_cb, NULL);

	/* Initialize vibration_handle (Use vibration now) */
	cur_h_data.handle = INVALID_HANDLE;
	cur_h_data.priority = PRIORITY_MIN;
}

void haptic_exit(void)
{
	struct haptic_ops *ops;
	dd_list *elem;
	int r;

	/* remove watch */
	if (haptic_conf.sound_capture)
		vconf_ignore_key_changed(VCONFKEY_RECORDER_STATE, sound_capturing_cb);

	/* unregister notifier for below each event */
	unsubscribe_dbus_signal(NULL, id_sig_change_hardkey);
	unsubscribe_dbus_signal(NULL, id_sig_pwr_off_state);

	/* release haptic data memory */
	safe_free(haptic_conf.level_arr);

	if (!CHECK_VALID_OPS(h_ops, r))
		return;

	/* haptic exit for feedbackd */
	haptic_internal_exit();

	/* release plugin */
	DD_LIST_FOREACH(h_head, elem, ops) {
		if (ops->is_valid && ops->is_valid()) {
			if (ops->release)
				ops->release();
			h_ops = NULL;
			break;
		}
	}
}

static int haptic_start(void)
{
	_I("start");
	haptic_disabled = false;
	return 0;
}

static int haptic_stop(void)
{
	_I("stop");
	haptic_disabled = true;
	return 0;
}
